#!/usr/bin/env python
from typing import List
import os, sys, re

# Use with Testcontroller device config file for making config copy
# Used for multi-channel device with (nearly) identical characteristics
# After development for one channel you can add copies for the rest with this program
# The program tries to make a copy based on the items below and will report problems if possible
# Only use the to-be-copied config statements as input.
# You can paste the output in your Testcontroller config file
# the new config is saved in the subdirectory "copy" with the name of the input file
# NOTE: Don't forget to update your #metadef definitions including prefix/postfix

# developed with Python 3.9

# CHANGE THIS AS REQUIRED
# you can only convert one page at a time
current_setup_page_name: str = 'Channel_1'
# this string is part of control names or in other controls, identifying the channel
current_setup_string: str = 'CH1'
# this string is current prefix for channel specific SCPI read/write commands
current_setup_SCPI_prefix: str = 'C1:'
# this string is current comment header
current_setup_comment_string: str = 'C H A N N E L 1'

# replacements
new_setup_page_name: str = 'Channel_2'
new_setup_string: str = 'CH2'
new_setup_SCPI_prefix: str = 'C2:'
new_setup_comment_string: str = 'C H A N N E L 2'
# replacement for channel color
new_page_color: str = '(230,230,30)'

# Control names are globally defined in Testcontroller; copying control names to different pages will use the same control name
# Issuing a :update: on page C1 will also trigger an update of the same control on page C2, C3...
# This takes time (device access) and will delay the setup window update; solution: make names unique when possible!
# This program supports prefix or postfix adding to make control names globally unique
# You have to choose a different string for every channel!
# Prefix could be C1_, C2_ ..., but postfix will not be shown if you do: ch1 = '', ch2='_', ch3='__' and ch4='___'
# This will right shift the colon on the control name one space position per underscore in the setup window
# When HKJ supports trailing space suppression on his displayed control names this will disappear ;-)
# After you added the new channel you can run the check_setup.py script to see if on other pages
# you need to update :update: and :enable: tags
# NOTE: program assumes no prefix or postfix string is in input configuration
special_prefix:str = ''
special_postfix:str = ''


def process_setup_tags(tags: List[str],	cs_items:List[str], output_list: List[str]):
	lb = '{'
	rb = '}'
	# change page name
	cs_items[3] = new_setup_page_name
	# check / change control name
	if new_setup_string in cs_items[2]:
		cs_items[2] = cs_items[2].replace(current_setup_string, '{' + new_setup_string + '}')
		print(f"Warning: found {new_setup_string} of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
	if current_setup_string in cs_items[2]:
		cs_items[2] = cs_items[2].replace(current_setup_string, new_setup_string)
	# update control name
	if cs_items[1]!='color':
		if special_prefix != '':
			cs_items[2] = special_prefix + cs_items[2]
		if special_postfix != '':
			cs_items[2] = cs_items[2] + special_postfix
	# #cmdSetup to output list
	output_list.append(' '.join(cs_items))
	#process tags
	for tag in tags:
		k_lower:str = ''
		k:str = ''
		v:str = ''
		if (re.search(r'\s', tag)):
			k, v = tag.split(' ',1)
			k_lower = k.lower()
			#cleanup comma vs space separation on :update:
			if (k_lower == ':update:'):
				v = v.replace(',',' ')
				v = ' '.join(v.split())
		else:
			k = tag
			k_lower = k.lower()
		if (k_lower==':read:' or k_lower==':write:'):
			if new_setup_SCPI_prefix in v:
				v = v.replace(new_setup_SCPI_prefix, '{' + new_setup_SCPI_prefix + '}')
				print(f"Warning: found {new_setup_SCPI_prefix} in :read: or :write: {cs_items[2]}; changed to {lb}{new_setup_SCPI_prefix}{rb}")
			if current_setup_SCPI_prefix in v:
				v = v.replace(current_setup_SCPI_prefix, new_setup_SCPI_prefix)
		elif k_lower==':enable:':
			if special_prefix != '':
				v = v.replace(current_setup_page_name + '.', current_setup_page_name + '.' + special_prefix)
			if special_postfix != '':
				replacement = "\\1" + special_postfix
				v = re.sub(r"(\.\w+)", replacement, v)
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in :enable: of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if new_setup_string in v:
				v = v.replace(current_setup_string, new_setup_string)
			if new_setup_page_name in v:
				v = v.replace(new_setup_page_name, '{' + new_setup_page_name + '}')
				print(f"Warning: found {new_setup_page_name} in :update: of {cs_items[2]}; changed to {lb}{new_setup_page_name}{rb}")
			if current_setup_page_name in v:
				v = v.replace(current_setup_page_name, new_setup_page_name)
		elif k_lower==':update:':
			if special_prefix != '':
				controls: List[str] = v.split()
				new_controls: List[str] = []
				for control in controls:
					control = special_prefix + control
					new_controls.append(control)
				v = ' '.join(new_controls)
			if special_postfix != '':
				controls: List[str] = v.split()
				new_controls: List[str] = []
				for control in controls:
					control = control + special_postfix
					new_controls.append(control)
				v = ' '.join(new_controls)
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in :update: of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if current_setup_string in v:
				v = v.replace(current_setup_string, new_setup_string)
		elif k_lower==':tip:':
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in :tip: of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if current_setup_string in v:
				v = v.replace(current_setup_string, new_setup_string)
		elif (re.match(r'^:', k)):
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in {k} of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if current_setup_string in v:
				v = v.replace(current_setup_string, '{' + current_setup_string + '}')
				print(f"Warning: found {current_setup_string} in {k} of {cs_items[2]}; changed to {lb}{current_setup_string}{rb}")
		elif (re.match(r'^;', k)):
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in comment of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if current_setup_string in v:
				v = v.replace(current_setup_string, new_setup_string)
		else:
			if new_setup_string in k:
				k = k.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in entry line of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if new_setup_string in v:
				v = v.replace(new_setup_string, '{' + new_setup_string + '}')
				print(f"Warning: found {new_setup_string} in entry line of {cs_items[2]}; changed to {lb}{new_setup_string}{rb}")
			if current_setup_string in k:
				k = k.replace(current_setup_string, new_setup_string)
			if current_setup_string in v:
				v = v.replace(current_setup_string, new_setup_string)
			cntrl = cs_items[1].lower()
			if cntrl=='selector':
				if special_prefix != '':
					v = v.replace(current_setup_page_name + '.', current_setup_page_name + '.' + special_prefix)
				if special_postfix != '':
					controls: List[str] = v.split()
					new_controls: List[str] = []
					for control in controls:
						control = control + special_postfix
						new_controls.append(control)
					v = ' '.join(new_controls)
				if new_setup_page_name in v:
					v = v.replace(new_setup_page_name, '{' + new_setup_page_name + '}')
					print(f"Warning: found {new_setup_page_name} in entry line of {cs_items[2]}; changed to {lb}{new_setup_page_name}{rb}")
				if current_setup_page_name in v:
					v = v.replace(current_setup_page_name, new_setup_page_name)
			if cntrl=='color':
				if re.search(r'\(',k):
					k = new_page_color
		# put result on output_list
		if v =='':
			output_list.append(k)
		else:
			output_list.append(k + ' ' + v)




def main():
	cfg_input: List[str] = []
	output_list: List[str] = []
	tag_lines: List[str] = []
	comment_lines: List[str] = []
	first_cmdsetup: bool = False
	cs_items: List[str] = []
	if (len(sys.argv) < 2):
		print('Usage: copy_setup.py input_config_file')
		sys.exit(1)
	if (not os.path.isfile(sys.argv[1])):
		print(f'Input file {sys.argv[1]} does not exist.')
		sys.exit(1)
	input_match = re.match(r"(.*[\\\/]|)(.*).txt$",sys.argv[1], re.IGNORECASE) 
	if (not input_match):
		print(f'Input file {sys.argv[1]} is not a .txt file.')
		sys.exit(1)
	[input_dir_name, input_file_name] = input_match.group(1, 2)
	new_dir_name = input_dir_name + 'copy'
	if (not os.path.isdir(new_dir_name)):
		os.mkdir(new_dir_name)
	filename = new_dir_name + '/' + input_file_name
	# open output files
	with open(f"{filename}.txt", 'w', encoding='UTF8') as fo:
		# read config file
		with open(sys.argv[1], encoding='utf-8') as f:
			try:
				cfg_input = f.readlines()
			except:
				print(f'Could not process input file {sys.argv[1]}')
			# check config file
			line_count: int = 0
			for line in cfg_input:
				line_count += 1
				line = line.strip()
				if (first_cmdsetup == False):
					# first #cmdsetup command found
					if (re.match(r'^#cmdsetup ', line, re.I)):
						# get cs_items on: #cmdsetup  control-type control-name page-name
						#             	> cs_items[0] cs_items[1]  cs_items[2]  cs_items[3]  
						cs_items = line.split()
						if len(cs_items) == 3:
							print(f"Error: no page name on {line}; program requires a page name")
						if len(cs_items) == 4:
							if cs_items[3]==current_setup_page_name:
								first_cmdsetup = True
							else:
								print(f"Error: wrong page name on {line}; program requires {current_setup_page_name}")
						else:
							print(f"Error: {line} is invalid; program requires a 4 item format")
					# everything before first #cmdsetup command
					else:
						if  re.match(r';', line):
							if current_setup_page_name in line:
								line = line.replace(current_setup_page_name, new_setup_page_name)
							if current_setup_SCPI_prefix in line:
								line = line.replace(current_setup_SCPI_prefix, new_setup_SCPI_prefix)
							if current_setup_string in line:
								line = line.replace(current_setup_string, new_setup_string)
							if current_setup_comment_string in line:
								line = line.replace(current_setup_comment_string, new_setup_comment_string)
						output_list.append(line)
				# next #cmdsetup found
				else:
					if (re.match(r'#cmdsetup', line, re.I)):
						process_setup_tags(tag_lines, cs_items, output_list)
						for c in comment_lines:
							output_list.append(c)
						cs_items = []
						cs_items = line.split()
						if len(cs_items) == 3:
							print(f"Error: no page name on {line}; program requires a page name")
						if len(cs_items) == 4:
							if cs_items[3]!=current_setup_page_name:
								print(f"Error: wrong page name on {line}; program requires {current_setup_page_name}")
						else:
							print(f"Error: {line} is invalid; program requires a 4 item format")
						tag_lines = []
						comment_lines = []
					# empty lines
					elif re.match(r'\s*$', line):
						comment_lines.append(line)
					# and comment lines
					elif  re.match(r';', line):
						if current_setup_page_name in line:
							line = line.replace(current_setup_page_name, new_setup_page_name)
						if current_setup_SCPI_prefix in line:
							line = line.replace(current_setup_SCPI_prefix, new_setup_SCPI_prefix)
						if current_setup_string in line:
							line = line.replace(current_setup_string, new_setup_string)
						if current_setup_comment_string in line:
							line = line.replace(current_setup_comment_string, new_setup_comment_string)
						comment_lines.append(line)
					# tags & other # cmdsetup lines found
					else:
						tag_lines.append(line)
						# comment lines in between #cmdsetup lines are saved
						# empty lines are omitted 
						if (len(comment_lines) > 0):
							for c in comment_lines:
								if (re.match(r';', c)):
									tag_lines.append(c)
								comment_lines = []
				# don't forget the last #cmdsetup
				if (line_count == len(cfg_input)):
						process_setup_tags(tag_lines, cs_items, output_list)
						for c in comment_lines:
							output_list.append(c)
			# create content of output files
			for entry in output_list:
					fo.write(f"{entry}\n")
			
if __name__ == '__main__':
	main()
 